// votesim.c 2013-12-11 Claude Heiland-Allen <claude@mathr.co.uk>
// http://mathr.co.uk/blog/2013-12-11_distance_estimation_for_voting_simulation_visualisation.html
//
// gcc -std=c99 -Wall -pedantic -Wextra -lm -lcairo -O3 votesim.c 
// ./a.out
// ctrl-c when bored, but should eventually terminate anyway

#include <cairo/cairo.h>
#include <complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define candidates 4
const complex double candidate[candidates] =
  {  84 + 114 * I
  ,  36 + 108 * I
  ,  14 + 124 * I
  , 186 + 140 * I
  };
complex double bisectors[candidates][candidates][2];

#define electorate 10000
const double sigma = 100;

int votes[candidates];
double radii[electorate];

const int width = 200;
const int height = 200;
const int scale = 10;

complex double normal() {
  double u = (rand() + 1) / (double) RAND_MAX;
  double v = (rand() + 1) / (double) RAND_MAX;
  double r = sqrt(- 2 * log(u));
  double t = 6.283185307179586 * v;
  return r * (cos(t) + I * sin(t));
}

int cmp_double(const void *x, const void *y) {
  const double *a = x;
  const double *b = y;
  if (*a < *b) return -1;
  if (*a > *b) return  1;
  return 0;
}

int cmp_int(const void *x, const void *y) {
  const int *a = x;
  const int *b = y;
  if (*a < *b) return -1;
  if (*a > *b) return  1;
  return 0;
}

double distance_from_bisector(complex double p, int i, int j) {
  complex double m = bisectors[i][j][0];
  complex double d = bisectors[i][j][1];
  complex double pm = p - m;
  double u = creal(pm) * creal(d) + cimag(pm) * cimag(d);
  complex double n = m + u * d;
  return cabs(p - n);
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;
  for (int i = 0; i < candidates; ++i) {
    for (int j = i + 1; j < candidates; ++j) {
      bisectors[i][j][0] = (candidate[i] + candidate[j]) / 2;
      bisectors[i][j][1] = I * (candidate[i] - candidate[j]);
      bisectors[i][j][1] /= cabs(bisectors[i][j][1]);
    }
  }
  cairo_surface_t *image = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, scale * width, scale * height);
  unsigned char *data = cairo_image_surface_get_data(image);
  int stride = cairo_image_surface_get_stride(image);
  int channels = 4;
  memset(data, 0, scale * height * stride);
  cairo_t *cr = cairo_create(image);
  cairo_surface_mark_dirty(image);
  int count = 0;
  int counter = 0;
  int ocounter = 0;
  do {
    time_t t1 = time(0);
    do {
      int cx = rand() % (scale * width);
      int cy = rand() % (scale * height);
      if (data[cy * stride + cx * channels]) continue;
      memset(votes, 0, candidates * sizeof(int));
      for (int e = 0; e < electorate; ++e) {
        complex double voter = (cx + cy * I) / scale + sigma * normal();
        double distance = 1.0 / 0.0;
        int vote = -1;
        for (int c = 0; c < candidates; ++c) {
          double d = cabs(voter - candidate[c]);
          if (d < distance) {
            distance = d;
            vote = c;
          }
        }
        votes[vote]++;
        double radius = 1.0 / 0.0;
        for (int i = 0; i < candidates; ++i) {
          for (int j = i + 1; j < candidates; ++j) {
            double r = distance_from_bisector(voter, i, j);
            if (r < radius) {
              radius = r;
            }
          }
        }
        radii[e] = radius;
      }
      int winner = -1;
      int winvotes = 0;
      for (int c = 0; c < candidates; ++c) {
        if (winvotes < votes[c]) {
          winvotes = votes[c];
          winner = c;
        }
      }
      qsort(votes, candidates, sizeof(int), cmp_int);
      qsort(radii, electorate, sizeof(double), cmp_double);
      int swing = (votes[candidates - 1] - votes[candidates - 2]) / 2;
      double swing_radius = radii[swing];
      double k = (winner + 0.5) / candidates;
      cairo_set_source_rgba(cr, k, k, k, 1);
      cairo_arc(cr, cx, cy, scale * swing_radius, 0, 6.283185307179586);
      cairo_close_path(cr);
      cairo_fill(cr);
      cairo_surface_flush(image);
      counter++;
    } while (time(0) - t1 < 5);
    if (counter == ocounter) break;
    ocounter = counter;
    char filename[64];
    snprintf(filename, 64, "%02d-%08d.png", count++, counter);
    cairo_surface_write_to_png(image, filename);
  } while (1);
  return 0;
}
